var com_commandfusion_lifx = function(instanceID, instanceSettings, moduleDef) {
	/* ----------------------------------------------------
	 * MODULE PRIVATE MEMBERS
	 * ----------------------------------------------------
	 */
	var moduleName = "LIFX";
	var settingsURL = "";
    var port =  56700;
    var headerLength =  36;
    var protocol = 0x3400;
    var systemName = "LIFX";
    var systemNameInstance;
    var systemNameBroadcast = "LIFX_BROADCAST";
    var systemNameBroadcastInstance;
    var feedbackName = "LIFX_FB";
    var feedbackNameBroadcast = "LIFX_BROADCAST_FB";
    var gateways = [];
    var bulbs = [];
    var receivedBytes = [];
    var currentZoneID, currentZoneIndex = 0;
    var globalDelay = 3000; // delay before doing manual status requests after any power/dim commands
    var currentBulb = null;
	var projectData = {
		token: "",
		serverURL: "http://cc.commandfusion.com",
		settingsSaveURL: "",
		logEventURL: "",
		zones: {}
	};
    var settings = {
          "bulbZones": {} // mapping bulbs by their MAC address to zone IDs bulbZones[mac] = zoneID;
    };
	var listTag = "l" + instanceID + "_lights";
	var configZoneList = "l" + instanceID + "_config_zones";
	var configBulbList = "l" + instanceID + "_config_lights";
	var colorStringOff = "rgba(0,0,0,0.3)";

	/* ----------------------------------------------------
	 * INTERNAL FUNCTIONS
	 * ----------------------------------------------------
	 */
    function makeReadable(bytes) {
        var readable = "", i;
        for (i = 0; i < bytes.length; i++) {
            readable += "\\x" + ("0"+bytes.charCodeAt(i).toString(16).toUpperCase()).slice(-2);
        }
        return readable;
    }

	function makeReadableRaw(bytes) {
		var readable = "", i;
		for (i = 0; i < bytes.length; i++) {
			readable += ("0"+bytes.charCodeAt(i).toString(16).toUpperCase()).slice(-2);
		}
		return readable;
	}

	function makeUnreadable(str) {
		var unreadable = "", i;
		for (i = 0; i < str.length; i+=2) {
			unreadable += String.fromCharCode(parseInt(str.substr(i, 2), 16))
		}
		return unreadable;
	}

    function rgb2hsv () {
        var rr, gg, bb,
            r = arguments[0] / 255,
            g = arguments[1] / 255,
            b = arguments[2] / 255,
            h, s,
            v = Math.max(r, g, b),
            diff = v - Math.min(r, g, b),
            diffc = function(c){
                return (v - c) / 6 / diff + 1 / 2;
            };

        if (diff == 0) {
            h = s = 0;
        } else {
            s = diff / v;
            rr = diffc(r);
            gg = diffc(g);
            bb = diffc(b);

            if (r === v) {
                h = bb - gg;
            }else if (g === v) {
                h = (1 / 3) + rr - bb;
            }else if (b === v) {
                h = (2 / 3) + gg - rr;
            }
            if (h < 0) {
                h += 1;
            }else if (h > 1) {
                h -= 1;
            }
        }
        return {
            h: Math.round(h * 65535),
            s: Math.round(s * 65535),
            v: Math.round(v * 65535)
        };
    }

    function hsv2rgb (h, s, v) {
        /**
         * HSV to RGB color conversion
         *
         * H runs from 0 to 360 degrees
         * S and V run from 0 to 1
         *
         * Ported from the excellent java algorithm by Eugene Vishnevsky at:
         * http://www.cs.rit.edu/~ncs/color/t_convert.html
         *
         * Returns object with params in 0-255 range.
         */
        var r, g, b;
        var i;
        var f, p, q, t;

        // For LIFX, HSV are all represented as 16bit values (2 bytes) 0-65535 range.
        // This function requires S and V to be in 0-1 range, and H to be in 0-360 range, so convert.
        h = (360 / 65535) * h;
        s = s / 65535;
        v = v / 65535;

        if (s == 0) {
            // Achromatic (grey)
            r = g = b = v;
            return {r: Math.round(r * 255), g: Math.round(g * 255), b: Math.round(b * 255)};
        }

        h /= 60; // sector 0 to 5
        i = Math.floor(h);
        f = h - i; // factorial part of h
        p = v * (1 - s);
        q = v * (1 - s * f);
        t = v * (1 - s * (1 - f));

        switch(i) {
            case 0:
                r = v;
                g = t;
                b = p;
                break;
            case 1:
                r = q;
                g = v;
                b = p;
                break;
            case 2:
                r = p;
                g = v;
                b = t;
                break;
            case 3:
                r = p;
                g = q;
                b = v;
                break;
            case 4:
                r = t;
                g = p;
                b = v;
                break;
            default: // case 5:
                r = v;
                g = p;
                b = q;
        }

        r= Math.round(r * 255);
        //r= r.toString(16);

        g= Math.round(g * 255);
        //g= g.toString(16);

        b= Math.round(b * 255);
        //b= b.toString(16);
        return {r: r, g: g, b: b};
    }

    function hsv2hsl (h, s, v) {

        // Convert from H from 0-65535 range to 0-360 range
        h = (360 / 65535) * h;
        // Convert from S and V from 0-65535 range to 0-1 range
        s = (1 / 65535) * s;
        v = (1 / 65535) * v;

        // convert to HSL
        var l = (2 - s) * v;
        s = s * v;
        s /= (l <= 1) ? l : 2 - l;
        l /= 2;

        // correct a division-by-zero error
        if (isNaN(s)) s = 0;

        var hsl = {'h' : h, 's' : s, 'l' : l };

        CF.log("h: " + hsl.h + ", s: " + hsl.s + ", l:" + hsl.l);

        // Convert back to 0-65535 range
        hsl.h = (hsl.h / 360) * 65535;
        hsl.s = hsl.s * 65535;
        hsl.l = hsl.l * 65535;

        CF.log("h: " + hsl.h + ", s: " + hsl.s + ", l:" + hsl.l);

        return hsl;
    }

    function hsl2hsv (h, s, l) {

        //CF.log("h: " + h + ", s: " + s + ", l:" + l);

        // Convert from H from 0-65535 range to 0-360 range
        h = (360 / 65535) * h;
        // Convert from S and V from 0-65535 range to 0-1 range
        s = (1 / 65535) * s;
        l = (1 / 65535) * l;

        l *= 2;
        s *= (l <= 1) ? l : 2 - l;

        // create the HSV components
        var hsv = {'h' : 0, 's' : 0, 'v' : 0};

        hsv.h = h;
        hsv.v = (l + s) / 2;
        hsv.s = (2 * s) / (l + s)

//        // set a temporary value
//        var t = s * (l < 50 ? l : 100 - l) / 100;
//
//        // store the HSV components
//        var hsv =
//        {
//            'h' : h,
//            's' : 200 * t / (l + t),
//            'v' : t + l
//        };

        // Convert back to 0-65535 range
        hsv.h = (hsv.h / 360) * 65535;
        hsv.s = hsv.s * 65535;
        hsv.v = hsv.v * 65535;

        return hsv;
    }

    function toPercent(level) {
        return Math.round((100/65535) * level) + "%";
    }

    function pack(format) {
        //  discuss at: http://phpjs.org/functions/pack/
        // original by: Tim de Koning (http://www.kingsquare.nl)
        //    parts by: Jonas Raoni Soares Silva (http://www.jsfromhell.com)
        // bugfixed by: Tim de Koning (http://www.kingsquare.nl)
        //        note: Float encoding by: Jonas Raoni Soares Silva
        //        note: Home: http://www.kingsquare.nl/blog/12-12-2009/13507444
        //        note: Feedback: phpjs-pack@kingsquare.nl
        //        note: 'machine dependent byte order and size' aren't
        //        note: applicable for JavaScript; pack works as on a 32bit,
        //        note: little endian machine
        //   example 1: pack('nvc*', 0x1234, 0x5678, 65, 66);
        //   returns 1: '4xVAB'
        //   example 2: pack('H4', '2345')
        //   returns 2: '#E'
        //   example 3: pack('H*', 'D5')
        //   returns 3: 'Õ'
        //   example 4: pack('d', -100.876)
        //   returns 4: "\u0000\u0000\u0000\u0000\u00008YÀ"

        var formatPointer = 0,
            argumentPointer = 1,
            result = '',
            argument = '',
            i = 0,
            r = [],
            instruction, quantifier, word, precisionBits, exponentBits, extraNullCount;

        // vars used by float encoding
        var bias, minExp, maxExp, minUnnormExp, status, exp, len, bin, signal, n, intPart, floatPart, lastBit, rounded, j,
            k, tmpResult;

        while (formatPointer < format.length) {
            instruction = format.charAt(formatPointer);
            quantifier = '';
            formatPointer++;
            while ((formatPointer < format.length) && (format.charAt(formatPointer)
                .match(/[\d\*]/) !== null)) {
                quantifier += format.charAt(formatPointer);
                formatPointer++;
            }
            if (quantifier === '') {
                quantifier = '1';
            }

            // Now pack variables: 'quantifier' times 'instruction'
            switch (instruction) {
                case 'a':
                // NUL-padded string
                case 'A':
                    // SPACE-padded string
                    if (typeof arguments[argumentPointer] === 'undefined') {
                        throw new Error('Warning:  pack() Type ' + instruction + ': not enough arguments');
                    } else {
                        argument = String(arguments[argumentPointer]);
                    }
                    if (quantifier === '*') {
                        quantifier = argument.length;
                    }
                    for (i = 0; i < quantifier; i++) {
                        if (typeof argument[i] === 'undefined') {
                            if (instruction === 'a') {
                                result += String.fromCharCode(0);
                            } else {
                                result += ' ';
                            }
                        } else {
                            result += argument[i];
                        }
                    }
                    argumentPointer++;
                    break;
                case 'h':
                // Hex string, low nibble first
                case 'H':
                    // Hex string, high nibble first
                    if (typeof arguments[argumentPointer] === 'undefined') {
                        throw new Error('Warning: pack() Type ' + instruction + ': not enough arguments');
                    } else {
                        argument = arguments[argumentPointer];
                    }
                    if (quantifier === '*') {
                        quantifier = argument.length;
                    }
                    if (quantifier > argument.length) {
                        throw new Error('Warning: pack() Type ' + instruction + ': not enough characters in string');
                    }

                    for (i = 0; i < quantifier; i += 2) {
                        // Always get per 2 bytes...
                        word = argument[i];
                        if (((i + 1) >= quantifier) || typeof argument[i + 1] === 'undefined') {
                            word += '0';
                        } else {
                            word += argument[i + 1];
                        }
                        // The fastest way to reverse?
                        if (instruction === 'h') {
                            word = word[1] + word[0];
                        }
                        result += String.fromCharCode(parseInt(word, 16));
                    }
                    argumentPointer++;
                    break;

                case 'c':
                // signed char
                case 'C':
                    // unsigned char
                    // c and C is the same in pack
                    if (quantifier === '*') {
                        quantifier = arguments.length - argumentPointer;
                    }
                    if (quantifier > (arguments.length - argumentPointer)) {
                        throw new Error('Warning:  pack() Type ' + instruction + ': too few arguments');
                    }

                    for (i = 0; i < quantifier; i++) {
                        result += String.fromCharCode(arguments[argumentPointer]);
                        argumentPointer++;
                    }
                    break;

                case 's':
                // signed short (always 16 bit, machine byte order)
                case 'S':
                // unsigned short (always 16 bit, machine byte order)
                case 'v':
                    // s and S is the same in pack
                    if (quantifier === '*') {
                        quantifier = arguments.length - argumentPointer;
                    }
                    if (quantifier > (arguments.length - argumentPointer)) {
                        throw new Error('Warning:  pack() Type ' + instruction + ': too few arguments');
                    }

                    for (i = 0; i < quantifier; i++) {
                        result += String.fromCharCode(arguments[argumentPointer] & 0xFF);
                        result += String.fromCharCode(arguments[argumentPointer] >> 8 & 0xFF);
                        argumentPointer++;
                    }
                    break;

                case 'n':
                    // unsigned short (always 16 bit, big endian byte order)
                    if (quantifier === '*') {
                        quantifier = arguments.length - argumentPointer;
                    }
                    if (quantifier > (arguments.length - argumentPointer)) {
                        throw new Error('Warning: pack() Type ' + instruction + ': too few arguments');
                    }

                    for (i = 0; i < quantifier; i++) {
                        result += String.fromCharCode(arguments[argumentPointer] >> 8 & 0xFF);
                        result += String.fromCharCode(arguments[argumentPointer] & 0xFF);
                        argumentPointer++;
                    }
                    break;

                case 'i':
                // signed integer (machine dependent size and byte order)
                case 'I':
                // unsigned integer (machine dependent size and byte order)
                case 'l':
                // signed long (always 32 bit, machine byte order)
                case 'L':
                // unsigned long (always 32 bit, machine byte order)
                case 'V':
                    // unsigned long (always 32 bit, little endian byte order)
                    if (quantifier === '*') {
                        quantifier = arguments.length - argumentPointer;
                    }
                    if (quantifier > (arguments.length - argumentPointer)) {
                        throw new Error('Warning:  pack() Type ' + instruction + ': too few arguments');
                    }

                    for (i = 0; i < quantifier; i++) {
                        result += String.fromCharCode(arguments[argumentPointer] & 0xFF);
                        result += String.fromCharCode(arguments[argumentPointer] >> 8 & 0xFF);
                        result += String.fromCharCode(arguments[argumentPointer] >> 16 & 0xFF);
                        result += String.fromCharCode(arguments[argumentPointer] >> 24 & 0xFF);
                        argumentPointer++;
                    }

                    break;
                case 'N':
                    // unsigned long (always 32 bit, big endian byte order)
                    if (quantifier === '*') {
                        quantifier = arguments.length - argumentPointer;
                    }
                    if (quantifier > (arguments.length - argumentPointer)) {
                        throw new Error('Warning:  pack() Type ' + instruction + ': too few arguments');
                    }

                    for (i = 0; i < quantifier; i++) {
                        result += String.fromCharCode(arguments[argumentPointer] >> 24 & 0xFF);
                        result += String.fromCharCode(arguments[argumentPointer] >> 16 & 0xFF);
                        result += String.fromCharCode(arguments[argumentPointer] >> 8 & 0xFF);
                        result += String.fromCharCode(arguments[argumentPointer] & 0xFF);
                        argumentPointer++;
                    }
                    break;

                case 'f':
                // float (machine dependent size and representation)
                case 'd':
                    // double (machine dependent size and representation)
                    // version original by IEEE754
                    precisionBits = 23;
                    exponentBits = 8;
                    if (instruction === 'd') {
                        precisionBits = 52;
                        exponentBits = 11;
                    }

                    if (quantifier === '*') {
                        quantifier = arguments.length - argumentPointer;
                    }
                    if (quantifier > (arguments.length - argumentPointer)) {
                        throw new Error('Warning:  pack() Type ' + instruction + ': too few arguments');
                    }
                    for (i = 0; i < quantifier; i++) {
                        argument = arguments[argumentPointer];
                        bias = Math.pow(2, exponentBits - 1) - 1;
                        minExp = -bias + 1;
                        maxExp = bias;
                        minUnnormExp = minExp - precisionBits;
                        status = isNaN(n = parseFloat(argument)) || n === -Infinity || n === +Infinity ? n : 0;
                        exp = 0;
                        len = 2 * bias + 1 + precisionBits + 3;
                        bin = new Array(len);
                        signal = (n = status !== 0 ? 0 : n) < 0;
                        n = Math.abs(n);
                        intPart = Math.floor(n);
                        floatPart = n - intPart;

                        for (k = len; k;) {
                            bin[--k] = 0;
                        }
                        for (k = bias + 2; intPart && k;) {
                            bin[--k] = intPart % 2;
                            intPart = Math.floor(intPart / 2);
                        }
                        for (k = bias + 1; floatPart > 0 && k; --floatPart) {
                            (bin[++k] = ((floatPart *= 2) >= 1) - 0);
                        }
                        for (k = -1; ++k < len && !bin[k];) {}

                        if (bin[(lastBit = precisionBits - 1 + (k = (exp = bias + 1 - k) >= minExp && exp <= maxExp ? k + 1 :
                            bias + 1 - (exp = minExp - 1))) + 1]) {
                            if (!(rounded = bin[lastBit])) {
                                for (j = lastBit + 2; !rounded && j < len; rounded = bin[j++]) {}
                            }
                            for (j = lastBit + 1; rounded && --j >= 0;
                                 (bin[j] = !bin[j] - 0) && (rounded = 0)) {}
                        }

                        for (k = k - 2 < 0 ? -1 : k - 3; ++k < len && !bin[k];) {}

                        if ((exp = bias + 1 - k) >= minExp && exp <= maxExp) {
                            ++k;
                        } else {
                            if (exp < minExp) {
                                if (exp !== bias + 1 - len && exp < minUnnormExp) { /*"encodeFloat::float underflow" */ }
                                k = bias + 1 - (exp = minExp - 1);
                            }
                        }

                        if (intPart || status !== 0) {
                            exp = maxExp + 1;
                            k = bias + 2;
                            if (status === -Infinity) {
                                signal = 1;
                            } else if (isNaN(status)) {
                                bin[k] = 1;
                            }
                        }

                        n = Math.abs(exp + bias);
                        tmpResult = '';

                        for (j = exponentBits + 1; --j;) {
                            tmpResult = (n % 2) + tmpResult;
                            n = n >>= 1;
                        }

                        n = 0;
                        j = 0;
                        k = (tmpResult = (signal ? '1' : '0') + tmpResult + bin.slice(k, k + precisionBits)
                            .join(''))
                            .length;
                        r = [];

                        for (; k;) {
                            n += (1 << j) * tmpResult.charAt(--k);
                            if (j === 7) {
                                r[r.length] = String.fromCharCode(n);
                                n = 0;
                            }
                            j = (j + 1) % 8;
                        }

                        r[r.length] = n ? String.fromCharCode(n) : '';
                        result += r.join('');
                        argumentPointer++;
                    }
                    break;

                case 'x':
                    // NUL byte
                    if (quantifier === '*') {
                        throw new Error('Warning: pack(): Type x: \'*\' ignored');
                    }
                    for (i = 0; i < quantifier; i++) {
                        result += String.fromCharCode(0);
                    }
                    break;

                case 'X':
                    // Back up one byte
                    if (quantifier === '*') {
                        throw new Error('Warning: pack(): Type X: \'*\' ignored');
                    }
                    for (i = 0; i < quantifier; i++) {
                        if (result.length === 0) {
                            throw new Error('Warning: pack(): Type X:' + ' outside of string');
                        } else {
                            result = result.substring(0, result.length - 1);
                        }
                    }
                    break;

                case '@':
                    // NUL-fill to absolute position
                    if (quantifier === '*') {
                        throw new Error('Warning: pack(): Type X: \'*\' ignored');
                    }
                    if (quantifier > result.length) {
                        extraNullCount = quantifier - result.length;
                        for (i = 0; i < extraNullCount; i++) {
                            result += String.fromCharCode(0);
                        }
                    }
                    if (quantifier < result.length) {
                        result = result.substring(0, parseInt(quantifier, 10));
                    }
                    break;

                default:
                    throw new Error('Warning:  pack() Type ' + instruction + ': unknown format code');
            }
        }
        if (argumentPointer < arguments.length) {
            throw new Error('Warning: pack(): ' + (arguments.length - argumentPointer) + ' arguments unused');
        }

        return result;
    }

    function unpack(format, data) {
        // http://kevin.vanzonneveld.net
        // +   original by: Tim de Koning (http://www.kingsquare.nl)
        // +      parts by: Jonas Raoni Soares Silva - http://www.jsfromhell.com
        // +      parts by: Joshua Bell - http://cautionsingularityahead.blogspot.nl/
        // +
        // +   bugfixed by: marcuswestin
        // %        note 1: Float decoding by: Jonas Raoni Soares Silva
        // %        note 2: Home: http://www.kingsquare.nl/blog/22-12-2009/13650536
        // %        note 3: Feedback: phpjs-unpack@kingsquare.nl
        // %        note 4: 'machine dependant byte order and size' aren't
        // %        note 5: applicable for JavaScript unpack works as on a 32bit,
        // %        note 6: little endian machine
        // *     example 1: unpack('d', "\u0000\u0000\u0000\u0000\u00008YÀ");
        // *     returns 1: { "": -100.875 }

        var formatPointer = 0, dataPointer = 0, result = {}, instruction = '',
            quantifier = '', label = '', currentData = '', currentResult = '', i = 0, j = 0,
            word = '', fbits = 0, ebits = 0, dataByteLength = 0;

        // Used by float decoding - by Joshua Bell
        //http://cautionsingularityahead.blogspot.nl/2010/04/javascript-and-ieee754-redux.html
        var fromIEEE754 = function (bytes, ebits, fbits) {
            // Bytes to bits
            var bits = [];
            for (var i = bytes.length; i; i -= 1) {
                var byte = bytes[i - 1];
                for (var j = 8; j; j -= 1) {
                    bits.push(byte % 2 ? 1 : 0);
                    byte = byte >> 1;
                }
            }
            bits.reverse();
            var str = bits.join('');

            // Unpack sign, exponent, fraction
            var bias = (1 << (ebits - 1)) - 1;
            var s = parseInt(str.substring(0, 1), 2) ? -1 : 1;
            var e = parseInt(str.substring(1, 1 + ebits), 2);
            var f = parseInt(str.substring(1 + ebits), 2);

            // Produce number
            if (e === (1 << ebits) - 1) {
                return f !== 0 ? NaN : s * Infinity;
            }
            else if (e > 0) {
                return s * Math.pow(2, e - bias) * (1 + f / Math.pow(2, fbits));
            }
            else if (f !== 0) {
                return s * Math.pow(2, -(bias - 1)) * (f / Math.pow(2, fbits));
            }
            else {
                return s * 0;
            }
        }

        while (formatPointer < format.length) {
            instruction = format.charAt(formatPointer);

            // Start reading 'quantifier'
            quantifier = '';
            formatPointer++;
            while ((formatPointer < format.length) &&
                (format.charAt(formatPointer).match(/[\d\*]/) !== null)) {
                quantifier += format.charAt(formatPointer);
                formatPointer++;
            }
            if (quantifier === '') {
                quantifier = '1';
            }


            // Start reading label
            label = '';
            while ((formatPointer < format.length) &&
                (format.charAt(formatPointer) !== '/')) {
                label += format.charAt(formatPointer);
                formatPointer++;
            }
            if (format.charAt(formatPointer) === '/') {
                formatPointer++;
            }

            // Process given instruction
            switch (instruction) {
                case 'a': // NUL-padded string
                case 'A': // SPACE-padded string
                    if (quantifier === '*') {
                        quantifier = data.length - dataPointer;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }
                    currentData = data.substr(dataPointer, quantifier);
                    dataPointer += quantifier;

                    if (instruction === 'a') {
                        currentResult = currentData.replace(/\0+$/, '');
                    } else {
                        currentResult = currentData.replace(/ +$/, '');
                    }
                    result[label] = currentResult;
                    break;

                case 'h': // Hex string, low nibble first
                case 'H': // Hex string, high nibble first
                    if (quantifier === '*') {
                        quantifier = data.length - dataPointer;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }
                    currentData = data.substr(dataPointer, quantifier);
                    dataPointer += quantifier;

                    if (quantifier > currentData.length) {
                        throw new Error('Warning: unpack(): Type ' + instruction +
                            ': not enough input, need ' + quantifier);
                    }

                    currentResult = '';
                    for (i = 0; i < currentData.length; i++) {
                        word = currentData.charCodeAt(i).toString(16);
                        if (instruction === 'h') {
                            word = word[1] + word[0];
                        }
                        currentResult += word;
                    }
                    result[label] = currentResult;
                    break;

                case 'c': // signed char
                case 'C': // unsigned c
                    if (quantifier === '*') {
                        quantifier = data.length - dataPointer;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }

                    currentData = data.substr(dataPointer, quantifier);
                    dataPointer += quantifier;

                    for (i = 0; i < currentData.length; i++) {
                        currentResult = currentData.charCodeAt(i);
                        if ((instruction === 'c') && (currentResult >= 128)) {
                            currentResult -= 256;
                        }
                        result[label + (quantifier > 1 ?
                            (i + 1) :
                            '')] = currentResult;
                    }
                    break;

                case 'S': // unsigned short (always 16 bit, machine byte order)
                case 's': // signed short (always 16 bit, machine byte order)
                case 'v': // unsigned short (always 16 bit, little endian byte order)
                    if (quantifier === '*') {
                        quantifier = (data.length - dataPointer) / 2;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }

                    currentData = data.substr(dataPointer, quantifier * 2);
                    dataPointer += quantifier * 2;

                    for (i = 0; i < currentData.length; i += 2) {
                        // sum per word;
                        currentResult = ((currentData.charCodeAt(i + 1) & 0xFF) << 8) +
                            (currentData.charCodeAt(i) & 0xFF);
                        if ((instruction === 's') && (currentResult >= 32768)) {
                            currentResult -= 65536;
                        }
                        result[label + (quantifier > 1 ?
                            ((i / 2) + 1) :
                            '')] = currentResult;
                    }
                    break;

                case 'n': // unsigned short (always 16 bit, big endian byte order)
                    if (quantifier === '*') {
                        quantifier = (data.length - dataPointer) / 2;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }

                    currentData = data.substr(dataPointer, quantifier * 2);
                    dataPointer += quantifier * 2;

                    for (i = 0; i < currentData.length; i += 2) {
                        // sum per word;
                        currentResult = ((currentData.charCodeAt(i) & 0xFF) << 8) +
                            (currentData.charCodeAt(i + 1) & 0xFF);
                        result[label + (quantifier > 1 ?
                            ((i / 2) + 1) :
                            '')] = currentResult;
                    }
                    break;

                case 'i': // signed integer (machine dependent size and byte order)
                case 'I': // unsigned integer (machine dependent size & byte order)
                case 'l': // signed long (always 32 bit, machine byte order)
                case 'L': // unsigned long (always 32 bit, machine byte order)
                case 'V': // unsigned long (always 32 bit, little endian byte order)
                    if (quantifier === '*') {
                        quantifier = (data.length - dataPointer) / 4;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }

                    currentData = data.substr(dataPointer, quantifier * 4);
                    dataPointer += quantifier * 4;

                    for (i = 0; i < currentData.length; i += 4) {
                        currentResult =
                            ((currentData.charCodeAt(i + 3) & 0xFF) << 24) +
                            ((currentData.charCodeAt(i + 2) & 0xFF) << 16) +
                            ((currentData.charCodeAt(i + 1) & 0xFF) << 8) +
                            ((currentData.charCodeAt(i) & 0xFF));
                        result[label + (quantifier > 1 ?
                            ((i / 4) + 1) :
                            '')] = currentResult;
                    }

                    break;

                case 'N': // unsigned long (always 32 bit, little endian byte order)
                    if (quantifier === '*') {
                        quantifier = (data.length - dataPointer) / 4;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }

                    currentData = data.substr(dataPointer, quantifier * 4);
                    dataPointer += quantifier * 4;

                    for (i = 0; i < currentData.length; i += 4) {
                        currentResult =
                            ((currentData.charCodeAt(i) & 0xFF) << 24) +
                            ((currentData.charCodeAt(i + 1) & 0xFF) << 16) +
                            ((currentData.charCodeAt(i + 2) & 0xFF) << 8) +
                            ((currentData.charCodeAt(i + 3) & 0xFF));
                        result[label + (quantifier > 1 ?
                            ((i / 4) + 1) :
                            '')] = currentResult;
                    }

                    break;

                case 'f': //float
                case 'd': //double
                    ebits = 8;
                    fbits = (instruction === 'f') ? 23 : 52;
                    dataByteLength = 4;
                    if (instruction === 'd') {
                        ebits = 11;
                        dataByteLength = 8;
                    }

                    if (quantifier === '*') {
                        quantifier = (data.length - dataPointer) / dataByteLength;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }

                    currentData = data.substr(dataPointer, quantifier * dataByteLength);
                    dataPointer += quantifier * dataByteLength;

                    for (i = 0; i < currentData.length; i += dataByteLength) {
                        data = currentData.substr(i, dataByteLength);

                        bytes = [];
                        for (j = data.length - 1; j >= 0; --j) {
                            bytes.push(data.charCodeAt(j));
                        }
                        result[label + (quantifier > 1 ?
                            ((i / 4) + 1) :
                            '')] = fromIEEE754(bytes, ebits, fbits);
                    }

                    break;

                case 'x': // NUL byte
                case 'X': // Back up one byte
                case '@': // NUL byte
                    if (quantifier === '*') {
                        quantifier = data.length - dataPointer;
                    } else {
                        quantifier = parseInt(quantifier, 10);
                    }

                    if (quantifier > 0) {
                        if (instruction === 'X') {
                            dataPointer -= quantifier;
                        } else {
                            if (instruction === 'x') {
                                dataPointer += quantifier;
                            } else {
                                dataPointer = quantifier;
                            }
                        }
                    }
                    break;

                default:
                    throw new Error('Warning:  unpack() Type ' + instruction +
                        ': unknown format code');
            }
        }
        return result;
    }

    function updateZoneList() {
        if (typeof projectData.zones !== 'undefined') {
            CF.listRemove(configZoneList);
            var zones = [];
            for (var i in projectData.zones) {
                var zone = projectData.zones[i];
                zones.push({s1: zone.name, s2: getBulbCount(zone.zone_id), d1: { tokens: { "zone_name": zone.name, "zone_id": zone.zone_id}}});
            }
            CF.listAdd(configZoneList, zones);
        } else {
            CFM.log("NO ZONE LIST!");
        }
    }

	function startDiscovery() {
        //CF.log(moduleName + " START DISCOVERY...");

        systemNameInstance = instanceID + "." + systemName;
        systemNameBroadcastInstance = instanceID + "." + systemNameBroadcast;

        CF.watch(CF.FeedbackMatchedEvent, systemNameBroadcastInstance, feedbackNameBroadcast, incomingData);

		CF.setSystemProperties(systemNameInstance, {enabled: false});	        // disable direct system
		CF.setSystemProperties(systemNameBroadcastInstance, {enabled: true});	// enable broadcast system

        gateways = [];
        bulbs = [];
        CF.listRemove(listTag);

        sendBroadcastPacket(encodePacket(0x02));
        setTimeout(function() {
            sendBroadcastPacket(encodePacket(0x02));
        }, 250);
        setTimeout(function() {
            sendBroadcastPacket(encodePacket(0x02));
        }, 500);
        //CFM.triggerOutput([instanceID, "light-list"], null);
	}

    function encodePacket(type, payload, target_mac, useProtocol) {
        payload = payload || "";
        protocol = useProtocol || protocol;
        type = type || 0x02;
        target_mac = target_mac || "";
        var gateway_mac = "";
        if (gateways[0]) {
            gateway_mac = gateways[0].gateway_mac || "";
        }
        var timestamp = 0;
        return pack("vvNa6na6nNNvn", payload.length + headerLength, protocol, 0, target_mac, 0, gateway_mac, 0, timestamp, timestamp, type, 0) + payload;
    }

    function sendBroadcastPacket(data) {
        CF.send(systemNameBroadcastInstance, data);
        //CF.log("OUTGOING -> " + systemNameBroadcastInstance + ": " + makeReadable(data));
    }

    function sendPacket(data) {
        CF.send(systemNameInstance, data);
    }

    function gatewayGetTime() {
        sendBroadcastPacket(encodePacket(0x04));
    }

	function getBulbCount(zoneID) {
		var count = 0;
		for (var mac in settings.bulbZones) {
			if (settings.bulbZones[mac] == zoneID) {
				count++;
			}
		}
		return count;
	}

    // Get the bulb either by 'name' or by 'mac'
    function getBulb(identifyBy, val) {
        if (!identifyBy || !val) return currentBulb;
        var bulb = null;
        if (identifyBy.toLowerCase() == "name") {
            bulb = getBulbByName(val);
        } else {
            bulb = getBulbByMac(val);
        }
        return bulb;
    }

    function getBulbByMac(mac) {
        for (var idx in bulbs) {
            if (bulbs.hasOwnProperty(idx)) {
                var bulb = bulbs[idx];
                if (bulb.target_mac == mac) {
                    return bulb;
                }
            }
        }
    }

    function getBulbByName(name) {
        for (var idx in bulbs) {
            if (bulbs.hasOwnProperty(idx)) {
                var bulb = bulbs[idx];
                if (bulb && bulb.label && bulb.label.toLowerCase() == name.toLowerCase()) {
                    return bulb;
                }
            }
        }
    }

    function getBulbIndexByMac(bulb, callback) {
	    var mac = bulb;
	    if (typeof bulb != 'string' && !(bulb instanceof String)) {
		    mac = bulb.target_mac;
	    }
        CF.listContents(listTag, 0, 0, function(items) {
            for (var idx in items) {
                var item = items[idx];
                if (item.d1 && item.d1.tokens && item.d1.tokens.target_mac && item.d1.tokens.target_mac == mac) {
                    callback(idx, bulb);
                }
            }
        });
    }

    function getBulbIndexByName(name) {
        for (var idx in bulbs) {
            if (bulbs.hasOwnProperty(idx)) {
                var bulb = bulbs[idx];
                if (bulb && bulb.label && bulb.label.toLowerCase() == name.toLowerCase()) {
                    return idx;
                }
            }
        }
    }

    function triggerBulbOutput(bulb, outputName, value) {
        // Trigger the named output, for both the bulb's name and MAC address
        CFM.triggerOutput([instanceID, outputName, bulb.label], value.toString());
        CFM.triggerOutput([instanceID, outputName, bulb.target_mac], value.toString());
    }

    function decimalToHex(d) {
        var hex = Number(d).toString(16);
        hex = ("00"+ hex).slice(-2);
        return hex;
    }

	function getColorString(bulb) {
		return "rgb("+bulb.rgb.r+","+bulb.rgb.g+","+bulb.rgb.b+")";
	}

	function getColorStringHSL(hue, sat, bright) {
		return "hsl("+Math.floor(((360/65535)*hue))+","+toPercent(sat, 65535)+","+toPercent(bright, 65535)+")";
	}

	function toPercent(level, max) {
		max = max || 65535;
		return Math.round((100/max) * level) + "%";
	}

    var ColorRequest = function(index, r, g, b) {
      return {
          index: index,
          r: r,
          g: g,
          b: b
      }
    };

    function incomingData(fbName, incomingData) {
        //CF.log("INCOMING: " + makeReadable(incomingData));
        // Append data to the buffer (multiple packets can arrive in one frame, and protocol doesn't offer an EOM to split with, so have to manually buffer and remove based on first byte which defines the packet length)
        receivedBytes += incomingData;
        // Unpack the byte array
        while (receivedBytes.length > 0) {
            //CF.log("LIFX BUFFER: " + makeReadable(receivedBytes));
            var unpacked = unpack("vsize/vprotocol/Nreserved1/a6target_mac/nreserved2/a6gateway_mac/nreserved3/Nts1/Nts2/vtype/nreserved4", receivedBytes);
            //CF.logObject(unpacked);
            //CF.log(makeReadable(unpacked.target_mac));
            if (receivedBytes.length < unpacked.size) {
                CFM.log("BUFFER MISSING DATA: " + (unpacked.size - receivedBytes.length));
                break;
            }
            // Strip the content from receivedBytes based on size of message defined in first byte of data.
            var data = receivedBytes.substr(0, unpacked.size);
            receivedBytes = receivedBytes.substr(unpacked.size);

            //CF.log("Bytes Remaining: " + receivedBytes.length);
            switch_breaker:
                switch (unpacked.type) {
                    case 0x03 :
                        var d = unpack("x" + headerLength + "header/cservice/Vport", data);
                        //CF.logObject(d);

                        gateways.push(unpacked);
                        //CF.log(moduleName + " Found a gateway bulb - " + unpacked.gateway_mac);

                        // Check if the bulb has been discovered already
                        for (var idx in bulbs) {
                            var bulb = bulbs[idx];
                            if (bulb.target_mac == unpacked.target_mac) {
                                // Already discovered, break now to skip adding to the collection again.
                                break switch_breaker;
                            }
                        }
                        CFM.feedInput([instanceID, "bulb-get-state"], null);
                        break;
                    case 0x6b:
                        // Bulb status
                        var d = unpack("x" + headerLength + "header/vhue/vsat/vbright/vkelvin/vdim/npower/a32label/Ntags1/Ntags2", data);
                        //CF.logObject(d);

                        var bulb = getBulbByMac(unpacked.target_mac);
                        if (!bulb) {
                            // Bulb not yet found, add it to the array
                            bulb = unpacked;
                            bulb.power = d.power;
                            bulb.label = d.label;
                            bulb.hue = d.hue;
                            bulb.sat = d.sat;
                            bulb.bright = d.bright;
                            bulb.kelvin = d.kelvin;
                            bulb.dim = d.dim;

                            bulbs.push(bulb);
                            CF.listAdd(listTag, [
                                {
                                    s1: bulb.label,
                                    d1: {
                                        tokens: {
                                            target_mac: bulb.target_mac,
                                            label: bulb.label
                                        },
                                        value: bulb.power
                                    },
                                    d2: {
                                        tokens: {
                                            target_mac: bulb.target_mac,
                                            label: bulb.label
                                        },
                                        value: !bulb.power
                                    },
                                    a1: {
                                        tokens: {
                                            target_mac: bulb.target_mac,
                                            label: bulb.label
                                        }
                                    },
                                    d3: {
                                        tokens: {
                                            target_mac: bulb.target_mac,
                                            label: bulb.label
                                        }
                                    }
                                }
                            ]);
                        } else {
                            bulb.power = d.power;
                            bulb.label = d.label;
                            bulb.hue = d.hue;
                            bulb.sat = d.sat;
                            bulb.bright = d.bright;
                            bulb.kelvin = d.kelvin;
                            bulb.dim = d.dim;
                        }

                        bulb.rgb = hsv2rgb(bulb.hue, bulb.sat, bulb.bright);
                        bulb.hsl = hsv2hsl(bulb.hue, bulb.sat, bulb.bright);

                        //CF.logObject(bulb.hsl);

                        // Trigger bulb outputs, for both the bulb's name and MAC address
                        triggerBulbOutput(bulb, "bulb-power-on", bulb.power ? 1 : 0);
                        triggerBulbOutput(bulb, "bulb-power-off", bulb.power ? 0 : 1);
                        triggerBulbOutput(bulb, "bulb-dim-level", bulb.hsl.l);
                        triggerBulbOutput(bulb, "bulb-dim-percent", toPercent(bulb.hsl.l));
                        triggerBulbOutput(bulb, "bulb-rgb-json", JSON.stringify(bulb.rgb));
                        triggerBulbOutput(bulb, "bulb-red", bulb.rgb.r);
                        triggerBulbOutput(bulb, "bulb-green", bulb.rgb.g);
                        triggerBulbOutput(bulb, "bulb-blue", bulb.rgb.b);

	                    var lightData = {id: bulb.label, state: (bulb.power ? "1" : "0"), level: bulb.dim, r: bulb.rgb.r, g: bulb.rgb.g, b: bulb.rgb.b};
	                    CFM.triggerOutput([instanceID, "light-data"], JSON.stringify([lightData]));

                        getBulbIndexByMac(bulb, function(idx) {
                            if (typeof idx === "undefined") return;
                            // Update the list with the actual bulb label
                            CF.listUpdate(listTag, [
                                {
                                    index: idx,
                                    s1: bulb.label,
                                    d1: bulb.power,
                                    d2: !bulb.power,
                                    s2: toPercent(bulb.hsl.l),
                                    a1: bulb.hsl.l,
	                                s3: {
		                                properties: {
			                                backgroundColor: bulb.power ? getColorString(bulb) : colorStringOff
		                                }
	                                }
                                }
                            ]);
                        });
                        break;
                    case 0x16:
                        // Bulb power state
                        var d = unpack("x" + headerLength + "header/npower", data);
                        //CF.logObject(d);

                        var bulb = getBulbByMac(unpacked.target_mac);
                        if (!bulb) {
                            CF.log("0x6B: BULB NOT FOUND - " + makeReadable(unpacked.target_mac));
                            return;
                        }
                        bulb.power = d.power;

                        // Trigger bulb power state output, for both the bulb's name and MAC address
                        triggerBulbOutput(bulb, "bulb-power-on", bulb.power ? 1 : 0);
                        triggerBulbOutput(bulb, "bulb-power-off", bulb.power ? 0 : 1);

	                    var lightData = {id: bulb.label, state: (bulb.power ? "1" : "0")};
	                    CFM.triggerOutput([instanceID, "light-data"], JSON.stringify([lightData]));

                        getBulbIndexByMac(bulb, function(idx, bulb) {
                            if (typeof idx === "undefined") return;
                            // Update the list with the bulb power state
                            CF.listUpdate(listTag, [
                                {
                                    index: idx,
                                    d1: bulb.power,
                                    d2: !bulb.power,
	                                s3: {
		                                properties: {
			                                backgroundColor: bulb.power ? getColorString(bulb) : colorStringOff
		                                }
	                                }
                                }
                            ]);
                        });
                        break;
                    default:
                        //CF.log("IGNORED TYPE: " + unpacked.type)
                }
        }
    }

	return {

		onParameterChange: function(parameter, value) {
			CFM.log(moduleName + " onParameterChange " + parameter + " = " + value);

		},

		onStartup: function() {
			CFM.log(moduleName + " onStartup");
			startDiscovery();
            updateZoneList();
		},

		/* ----------------------------------------------------
		 * MODULE INPUTS
		 * ----------------------------------------------------
		 */
		Input_StartDiscovery: function(ep, data) {
			startDiscovery();
		},

		Input_AllOff: function(ep, data) {
			// Turn off all the bulbs
            for (var idx in bulbs) {
                if (bulbs.hasOwnProperty(idx)) {
                    var bulb = bulbs[idx];
                    this.Input_BulbOff([instanceID, 'bulb-off', 'mac'], bulb.target_mac);
                }
            }
		},

		Input_AllOn: function(ep, data) {
            for (var idx in bulbs) {
                if (bulbs.hasOwnProperty(idx)) {
                    var bulb = bulbs[idx];
                    this.Input_BulbOn([instanceID, 'bulb-on', 'mac'], bulb.target_mac);
                }
            }
		},

        Input_AllSetRGB : function(ep, r, g, b, fadeTime) {
            for (var idx in bulbs) {
                if (bulbs.hasOwnProperty(idx)) {
                    var bulb = bulbs[idx];
                    this.Input_BulbSetRGB(ep, bulb.target_mac, r, g, b, fadeTime);
                }
            }
        },

        Input_BulbOff : function(ep, target_mac_or_name) {
            var identifyBy = ep[2];
            var bulb = getBulb(identifyBy, target_mac_or_name);
            if (!bulb) {
                CFM.log("BulbOff: BULB NOT FOUND - " + target_mac_or_name);
                return;
            }
            bulb.power = 0;
            sendBroadcastPacket(encodePacket(0x15, pack("v", 0), bulb.target_mac));
	        sendBroadcastPacket(encodePacket(0x15, pack("v", 0), bulb.target_mac, 0x1400));

            // Trigger outputs manually, without relying on feedback which is not always received
            triggerBulbOutput(bulb, "bulb-power-on", bulb.power ? 1 : 0);
            triggerBulbOutput(bulb, "bulb-power-off", bulb.power ? 0 : 1);

	        var lightData = {id: bulb.label, state: "0"};
	        CFM.triggerOutput([instanceID, "light-data"], JSON.stringify([lightData]));

            getBulbIndexByMac(bulb, function(idx, bulb) {
                if (typeof idx === "undefined") return;
                // Update the list with the bulb power state
                CF.listUpdate(listTag, [
                    {
                        index: idx,
                        d1: bulb.power,
                        d2: !bulb.power
                    }
                ]);
            });

            // Force the bulb to reply with it's power state so other apps stay in sync
            // Use a delay to allow for the bulb state to have been updated
            setTimeout(function(that) {
                that.Output_GetBulbPowerState([instanceID, "", bulb.target_mac]);
            }, globalDelay, this);
        },

        Input_BulbOn : function(ep, target_mac_or_name) {
            var identifyBy = ep[2];
            var bulb = getBulb(identifyBy, target_mac_or_name);
            if (!bulb) {
                CFM.log("BulbOn: BULB NOT FOUND - " + target_mac_or_name);
                return;
            }
            bulb.power = 65535;
            sendBroadcastPacket(encodePacket(0x15, pack("v", 1), bulb.target_mac));
	        sendBroadcastPacket(encodePacket(0x15, pack("v", 1), bulb.target_mac, 0x1400));

	        var lightData = {id: bulb.label, state: "1"};
	        CFM.triggerOutput([instanceID, "light-data"], JSON.stringify([lightData]));

            // Trigger outputs manually, without relying on feedback which is not always received
            triggerBulbOutput(bulb, "bulb-power-on", bulb.power ? 1 : 0);
            triggerBulbOutput(bulb, "bulb-power-off", bulb.power ? 0 : 1);
            // Update the list with the bulb power state
            getBulbIndexByMac(bulb, function(idx, bulb) {
                if (typeof idx === "undefined") return;
                CF.listUpdate(listTag, [
                    {
                        index: idx,
                        d1: bulb.power,
                        d2: !bulb.power
                    }
                ]);
            });

            // Force the bulb to reply with it's power state so other apps stay in sync
            // Use a delay to allow for the bulb state to have been updated
            setTimeout(function(that) {
                that.Output_GetBulbPowerState([instanceID, "", bulb.target_mac]);
            }, globalDelay, this);
        },

        Input_BulbToggle : function(ep, target_mac_or_name) {
            var identifyBy = ep[2];
            var bulb = getBulb(identifyBy, target_mac_or_name);
            if (!bulb) {
                CF.log("bulbToggle: BULB NOT FOUND - " + target_mac_or_name);
                return;
            }

            if (bulb.power) {
                this.Input_BulbOff(ep, bulb.target_mac);
            } else {
                this.Input_BulbOn(ep, bulb.target_mac);
            }
        },

        Input_BulbSetState : function (ep, state) {
            var identifyBy = ep[2];
            var target_mac_or_name = ep[3];
            var bulb = getBulb(identifyBy, target_mac_or_name);
            if (!bulb) {
                CFM.log("BulbSetState: BULB NOT FOUND - " + target_mac_or_name);
                return;
            }
            if (state == 1) {
                this.Input_BulbOn(ep, target_mac_or_name);
            } else {
                this.Input_BulbOff(ep, target_mac_or_name);
            }
        },

        // ep indexes:
        // 0 = instance ID, 1 = input ID, 2 = access mode (mac or name), 3 = H, 4 = S, 5 = V, 6 = fade time
        // value (second function argument) = the mac address or name of the bulb to target, or empty to use the last bulb selected.
        // HSV are in 0-65535 range
        Input_BulbSetHSV : function (ep, target_mac_or_name) {
            var identifyBy = ep[2];
            var bulb = getBulb(identifyBy, target_mac_or_name);
            if (!bulb) {
                CFM.log("BulbSetColor: BULB NOT FOUND - " + target_mac_or_name);
                return;
            }
            // remove unnecessary indexes from endpoint to create function call params for 'pack'
            var params = ep.slice(3);
            // Add some params to start of array for the 'pack' function call
            params.unshift("cvvvvV", 0);
            // Insert a zero param for kelvin before the delay param
            params[6] = params[5]; // Copy delay index to the end
            params[5] = 0; // Set kelvin to 0
            // If saturation color is zero, LIFX uses kelvin, so make sure saturation is always at least 1
            params[3] = params[3] || 1;

            bulb.hue = params[2];
            bulb.sat = params[3];
            bulb.bright = params[4];
            bulb.rgb = hsv2rgb(bulb.hue, bulb.sat, bulb.bright);
            bulb.hsl = hsv2hsl(bulb.hue, bulb.sat, bulb.bright);

            //CF.logObject(params);
            // Ensure enough params are in the array to meet the pack call requirements ("cvvvvV")
            while (params.length < params[0].length) {
                params.push(null);
            }

            sendBroadcastPacket(encodePacket(0x66, pack.apply(this, params), bulb.target_mac));

	        var lightData = {id: bulb.label, state: bulb.power ? "1" : "0", r: bulb.rgb.r, g: bulb.rgb.g, b: bulb.rgb.b, level: bulb.hsl.l};
	        CFM.triggerOutput([instanceID, "light-data"], JSON.stringify([lightData]));

            getBulbIndexByMac(bulb, function(idx, bulb) {
                if (typeof idx === "undefined") return;
	            CF.listUpdate(listTag, [
		            {
			            index: idx,
			            s2: toPercent(bulb.hsl.l),
			            a1: bulb.hsl.l,
			            s3: {
				            properties: {
					            backgroundColor: bulb.power ? getColorString(bulb) : colorStringOff
				            }
			            }
		            }
	            ]);
            });
        },

		// ep indexes:
		// 0 = instance ID, 1 = input ID, 2 = access mode (mac or name), 3 = Hue, 4 = fade time
		// value (second function argument) = the mac address or name of the bulb to target, or empty to use the last bulb selected.
		// HSV are in 0-65535 range
		Input_BulbSetHue : function (ep, target_mac_or_name) {
			var identifyBy = ep[2];
			var bulb = getBulb(identifyBy, target_mac_or_name);
			if (!bulb) {
				CFM.log("BulbSetHue: BULB NOT FOUND - " + target_mac_or_name);
				return;
			}
			// Create array of data to pack for command
			// [0] = pack flags
			// [1] = 0
			// [2] = hue
			// [3] = sat (at least 1, otherwise LIFX will use kelvin value)
			// [4] = brightness
			// [5] = kelvin (always zero when setting a color)
			// [6] = delay
			var params = ["cvvvvV", 0, ep[3], bulb.sat || 1, bulb.bright, 0, ep[4] || 1000];

			bulb.hue = params[2];
			bulb.rgb = hsv2rgb(bulb.hue, bulb.sat, bulb.bright);
			bulb.hsl = hsv2hsl(bulb.hue, bulb.sat, bulb.bright);

			// CF.logObject(params);
			// Ensure enough params are in the array to meet the pack call requirements ("cvvvvV")
			while (params.length < params[0].length) {
				params.push(null);
			}

			sendBroadcastPacket(encodePacket(0x66, pack.apply(this, params), bulb.target_mac));

			CF.setProperties([
				{join: instanceID + "_picker_color", "backgroundColor": getColorStringHSL(bulb.hue, 65535, 65535)},
				{join: instanceID + "_picker_result", "backgroundColor": getColorStringHSL(bulb.hue, bulb.sat, 65535)}
			]);

			var lightData = {id: bulb.label, state: bulb.power ? "1" : "0", r: bulb.rgb.r, g: bulb.rgb.g, b: bulb.rgb.b};
			CFM.triggerOutput([instanceID, "light-data"], JSON.stringify([lightData]));

			getBulbIndexByMac(bulb, function(idx, bulb) {
				if (typeof idx === "undefined") return;
				CF.listUpdate(listTag, [
					{
						index: idx,
						s2: toPercent(bulb.hsl.l),
						a1: bulb.hsl.l,
						s3: {
							properties: {
								backgroundColor: bulb.power ? getColorString(bulb) : colorStringOff
							}
						}
					}
				]);
			});
		},

		// ep indexes:
		// 0 = instance ID, 1 = input ID, 2 = access mode (mac or name), 3 = Sat, 4 = fade time
		// value (second function argument) = the mac address or name of the bulb to target, or empty to use the last bulb selected.
		// HSV are in 0-65535 range
		Input_BulbSetSat : function (ep, target_mac_or_name) {
			var identifyBy = ep[2];
			var bulb = getBulb(identifyBy, target_mac_or_name);
			if (!bulb) {
				CFM.log("BulbSetHue: BULB NOT FOUND - " + target_mac_or_name);
				return
			}
			// Create array of data to pack for command
			// [0] = pack flags
			// [1] = 0
			// [2] = hue
			// [3] = sat (at least 1, otherwise LIFX will use kelvin value)
			// [4] = brightness
			// [5] = kelvin (always zero when setting a color)
			// [6] = delay
			var params = ["cvvvvV", 0, bulb.hue, ep[3] || 1, bulb.bright, 0, ep[4] || 1000];

			bulb.sat = params[3];
			bulb.rgb = hsv2rgb(bulb.hue, bulb.sat, bulb.bright);
			bulb.hsl = hsv2hsl(bulb.hue, bulb.sat, bulb.bright);

			// CF.logObject(params);
			// Ensure enough params are in the array to meet the pack call requirements ("cvvvvV")
			while (params.length < params[0].length) {
				params.push(null);
			}

			sendBroadcastPacket(encodePacket(0x66, pack.apply(this, params), bulb.target_mac));

			CF.setProperties([
				{join: instanceID + "_picker_result", "backgroundColor": getColorStringHSL(bulb.hue, bulb.sat, 65535)}
			]);

			var lightData = {id: bulb.label, state: bulb.power ? "1" : "0", r: bulb.rgb.r, g: bulb.rgb.g, b: bulb.rgb.b};
			CFM.triggerOutput([instanceID, "light-data"], JSON.stringify([lightData]));

			getBulbIndexByMac(bulb, function(idx, bulb) {
				if (typeof idx === "undefined") return;
				CF.listUpdate(listTag, [
					{
						index: idx,
						s2: toPercent(bulb.hsl.l),
						a1: bulb.hsl.l,
						s3: {
							properties: {
								backgroundColor: bulb.power ? getColorString(bulb) : colorStringOff
							}
						}
					}
				]);
			});
		},

        // ep indexes:
        // 0 = instance ID, 1 = input ID, 2 = R, 3 = G, 4 = B, 5 = fade time, 6 = access mode (mac or name)
        // value (second function argument) = the mac address or name of the bulb to target, or empty to use the last bulb selected.
        Input_BulbSetRGB : function (ep, target_mac_or_name) {
            var identifyBy = ep[6];
            var bulb = getBulb(identifyBy, target_mac_or_name);
            if (!bulb) {
                CFM.log("BulbSetRGB: BULB NOT FOUND - " + target_mac_or_name);
                return;
            }
            // remove unnecessary indexes from endpoint to create function call params for 'rgb2hsv'
            var colorParams = ep.slice(2, 5);
            var fadeTime = ep[5] || 1000;
            var hsv = rgb2hsv.apply(this, colorParams);
            // If saturation color is zero, LIFX uses kelvin, so make sure saturation is always at least 1
            hsv.s = hsv.s || 1;

            sendBroadcastPacket(encodePacket(0x66, pack("cvvvvV", 0, hsv.h, hsv.s, hsv.v, 0, fadeTime), bulb.target_mac));

            bulb.rgb = {r: colorParams[0], g: colorParams[1], b: colorParams[2]};
            bulb.hue = hsv.h;
            bulb.sat = hsv.s;
            bulb.bright = hsv.v;
            bulb.hsl = hsv2hsl(bulb.hue, bulb.sat, bulb.bright);

	        var lightData = {id: bulb.label, state: bulb.power ? "1" : "0", r: bulb.rgb.r, g: bulb.rgb.g, b: bulb.rgb.b};
	        CFM.triggerOutput([instanceID, "light-data"], JSON.stringify([lightData]));

            getBulbIndexByMac(bulb, function(idx, bulb) {
                if (typeof idx === "undefined") return;
	            CF.listUpdate(listTag, [
		            {
			            index: idx,
			            s2: toPercent(bulb.hsl.l),
			            a1: bulb.hsl.l,
			            s3: {
				            "properties": {
					            backgroundColor: bulb.power ? getColorString(bulb) : colorStringOff
				            }
			            }
		            }
	            ]);
            });
        },

        Input_BulbDim : function (ep, dimLevel) {
            var identifyBy = ep[2];
            var target_mac_or_name = ep[3];
            var bulb = getBulb(identifyBy, target_mac_or_name);
            if (!bulb) {
                CFM.log("BulbDim: BULB NOT FOUND - " + target_mac_or_name);
                return;
            }
            var fadeTime = ep[4];

            if (typeof fadeTime === "undefined") {
                fadeTime = 0;
            }

            // first calculate the current HSL from the HSV components
            bulb.hsl = hsv2hsl(bulb.hue, bulb.sat, bulb.bright);

            // Change the L (luminance)
            bulb.hsl.l = dimLevel;

            // Convert back to HSV
            var hsv = hsl2hsv(bulb.hsl.h, bulb.hsl.s, bulb.hsl.l);

            CF.log("Dim Level: " + dimLevel);

            this.Input_BulbSetHSV([instanceID, "bulb-set-color", "mac", hsv.h, hsv.s, hsv.v, fadeTime], bulb.target_mac);

            //sendBroadcastPacket(encodePacket(0x68, pack("vV", dimLevel, fadeTime), bulb.target_mac));

            // Update the list item
//            CF.listUpdate(listTag, [
//                {
//                    index: getBulbIndexByMac(bulb),
//                    s3: colorURL + "r=" + colorParams[0] + "&g=" + colorParams[1] + "&b=" + colorParams[2]
//                }
//            ]);
        },

        Input_BulbGetState: function (ep, target_mac_or_name) {
            if (!target_mac_or_name) {
                CF.log("BulbGetState: NO TARGET MAC");
            } else {
                var identifyBy = ep[2];
                var bulb = getBulb(identifyBy, target_mac_or_name);
                if (!bulb) {
                    CFM.log("BulbGetState: BULB NOT FOUND - " + target_mac_or_name);
                    return;
                }
                target_mac_or_name = bulb.target_mac;
            }
            sendBroadcastPacket(encodePacket(0x65, null, target_mac_or_name));
        },

        Input_BulbGetTime: function (ep, target_mac_or_name) {
            if (!target_mac_or_name) {
                CF.log("BulbGetTime: NO TARGET MAC");
            } else {
                var identifyBy = ep[2];
                var bulb = getBulb(identifyBy, target_mac_or_name);
                if (!bulb) {
                    CFM.log("BulbGetTime: BULB NOT FOUND - " + target_mac_or_name);
                    return;
                }
                target_mac_or_name = bulb.target_mac;
            }

            sendBroadcastPacket(encodePacket(0x04, null, target_mac_or_name));
        },

        Input_BulbSelect: function (ep, target_mac_or_name) {
            if (!target_mac_or_name) {
                CF.log("BulbSelect: NO TARGET MAC");
            } else {
                var identifyBy = ep[2];
                var bulb = getBulb(identifyBy, target_mac_or_name);
                if (!bulb) {
                    CFM.log("BulbSelect: BULB NOT FOUND - " + target_mac_or_name);
                    return;
                }
                currentBulb = bulb;

	            // Update color picker UI
	            CF.setProperties([
		            {join: instanceID + "_picker_color", "backgroundColor": getColorStringHSL(bulb.hue, 65535, 65535)},
		            {join: instanceID + "_picker_result", "backgroundColor": getColorStringHSL(bulb.hue, bulb.sat, 65535)}
	            ]);
            }
        },

		Input_LightChanges: function (ep, json) {
			var jsonObj = JSON.parse(json);
			if (!jsonObj) {
				CF.log("LightChanges: Data not in valid JSON format: " + json);
				return;
			}
			if (jsonObj.constructor !== Array) {
				jsonObj = [jsonObj];
			}
			for (var index in jsonObj) {
				var lightData = jsonObj[index];
				var id = lightData.id;
				if (!id) {
					CF.log("Error: Invalid Bulb Name: " + id);
					CFM.triggerOutput([instanceID, "msg"], "Error: Invalid Bulb Name: " + id);
					return;
				}
				if (lightData.hasOwnProperty('state')) {
					// Turning a light on or off
					this.Input_BulbSetState([instanceID, 'bulb-set-state', 'name', id], lightData.state ? 1 : 0);
				} else if (lightData.hasOwnProperty('level')) {
					// Turning a light up or down
					this.Input_BulbDim([instanceID, 'bulb-dim', 'name', id], lightData.level);
				} else {
					CF.log("Unknown Light Change:");
					CF.logObject(lightData);
				}

			}
		},

        Input_ShowSubpage: function (ep, subpageName) {
            CF.setProperties({join: instanceID + "_" + subpageName, opacity: 1}, 0, 0.3);
            CF.setProperties({join: subpageName, opacity: 1}, 0, 0.3);
        },

        Input_HideSubpage: function (ep, subpageName) {
            CF.setProperties({join: instanceID + "_" + subpageName, opacity: 0}, 0, 0.1);
            CF.setProperties({join: subpageName, opacity: 0}, 0, 0.1);
        },

        Input_ZoneChange: function(ep, zoneString) {
            var zone = JSON.parse(zoneString);
            CFM.log("Zone Selected: " + zone.name);

            // Clear bulb list for the zone
            CF.listRemove(listTag);
            // Add any bulbs in the selected zone, as well as any bulbs that are unallocated
            var zoneBulbs = [];
            for (var mac in settings.bulbZones) {
                if (settings.bulbZones[mac] == zone.zone_id || !settings.bulbZones[mac]) {
                    var bulb = getBulbByMac(mac);
                    if (!bulb) continue;
                    zoneBulbs.push({
                        s1: bulb.label,
                        s2: toPercent(bulb.hsl.l),
                        a1: bulb.hsl.l,
                        d1: {
                            value: bulb.power,
                            tokens: {
                                target_mac: bulb.target_mac,
                                label: bulb.label
                            }
                        },
                        d2: {
                            value: !bulb.power,
                            tokens: {
                                target_mac: bulb.target_mac,
                                label: bulb.label
                            }
                        },
                        d3: {
                            tokens: {
                                target_mac: bulb.target_mac,
                                label: bulb.label
                            }
                        }
                    });
                }
            }
            CF.listAdd(listTag, zoneBulbs);
        },

		Config_SaveSettings: function(ep, settingsSaveURL) {
			projectData.settingsSaveURL = settingsSaveURL || projectData.settingsSaveURL;

			// Cannot use UTF-8 array keys for some reason - perhaps MySQL limitation?
			// So convert the mac address from byte data to ascii strings before saving.
			var storageSettings = {bulbZones: {}};
			for (var mac in settings.bulbZones) {
				storageSettings.bulbZones[makeReadableRaw(mac)] = settings.bulbZones[mac];
			}

			// Convert the module settings to JSON string
			var settingsString = JSON.stringify(storageSettings);

			// Save the settings into the local global token first
			CF.getJoin(CF.GlobalTokensJoin, function(j,v,tokens) {
				if (typeof tokens["module_settings"] !== "undefined") {
					// Read the current settings
					var settingsData = JSON.parse(tokens["module_settings"]);
					// Update just the settings for the current module
					settingsData[instanceID] = JSON.parse(settingsString);
					// Save the settings back to the global token
					CF.setToken(CF.GlobalTokensJoin, "module_settings", JSON.stringify(settingsData));
				}
			});

			// Save settings back to the cloud
			if (!settingsURL) {
				CF.log(moduleName + " SAVE SETTINGS: settings URL was not defined. Settings were not saved.");
				return;
			}
			CF.request(settingsURL, "POST", null, {'settings': settingsString, 'module_instance': instanceID}, function(status, headers, body) {
			});
		},

        Config_LoadSettings: function(ep, settingsData) {
            // Convert the settings data to an object
            try {
                var storageSettings = JSON.parse(settingsData);

	            // Convert ascii mac addresses back to raw bytes
	            settings = {bulbZones: {}};
	            for (var mac in storageSettings.bulbZones) {
		            settings.bulbZones[makeUnreadable(mac)] = storageSettings.bulbZones[mac];
	            }

                updateZoneList();
                CF.log(moduleName + " SETTINGS LOADED!");
            } catch (err) {
                CF.log(moduleName + " ERROR LOADING SETTINGS: " + err);
            }
        },

		Config_Startup: function(ep, projectInfo) {
			try {
				projectData = JSON.parse(projectInfo);
			} catch (e) {
				CF.log(moduleName + " ERROR PROCESSING PROJECT INFO: " + e.stack);
			}
		},

        Config_SelectZone: function(ep, tokenString) {
            var tokens = JSON.parse(tokenString);
            // Select the zone in the list
            CF.listUpdate(configZoneList, [{index: CF.AllItems, d1: 0}, {index: tokens["[item_index]"], d1: 1}]);
            currentZoneID = tokens["zone_id"];
	        currentZoneIndex = tokens["[item_index]"];
            // Update config bulb list to highlight bulbs that are already in the selected zone
            CF.listRemove(configBulbList);
            var zoneBulbs = [];
            for (var idx in bulbs) {
                var bulb = bulbs[idx];
                zoneBulbs.push({
                    s1: bulb.label,
                    d1: {
                        tokens: {
                            target_mac: bulb.target_mac
                        },
                        value: (typeof currentZoneID !== "undefined" && settings.bulbZones[bulb.target_mac] == currentZoneID) ? 1 : 0
                    }
                });
            }
            CF.listAdd(configBulbList, zoneBulbs);
        },

        Config_SelectBulb: function(ep, tokenString) {
            var tokens = JSON.parse(tokenString);
            CF.log("Bulb Selected: " + tokens["target_mac"]);
            // Toggle the bulb in the list
            var wasActive = (settings.bulbZones[tokens["target_mac"]] == currentZoneID);
            CF.listUpdate(configBulbList, [{index: tokens["[item_index]"], d1: !wasActive}]);
            // Add the bulb to the zone
            if (!wasActive) {
                settings.bulbZones[tokens["target_mac"]] = currentZoneID;
            } else {
                settings.bulbZones[tokens["target_mac"]] = undefined;
            }

	        CF.log(JSON.stringify(settings));

	        CF.listUpdate(configZoneList, [{index: currentZoneIndex, s2: getBulbCount(currentZoneID)}]);
        },

		/* ----------------------------------------------------
		 * MODULE OUTPUT GETTERS
		 * ----------------------------------------------------
		 */

        Output_GetBulbPowerState: function (ep) {
            var target_mac_or_name = ep[2];
            // Find the bulb by name
            var bulb = getBulbByName(target_mac_or_name);
            if (!bulb) {
                // Find the bulb by MAC address
                bulb = getBulbByMac(target_mac_or_name);
            }
            if (!bulb) {
                CFM.log("GetBulbPowerState: Bulb not found: " + target_mac_or_name);
                return;
            }

            // Request the actual state of the bulb, if we get a reply the real output will be triggered after the current stored state is returned below
            sendBroadcastPacket(encodePacket(0x14, null, bulb.target_mac));

            // Return the bulb power state as a zero or one.
            return bulb.power ? 1 : 0;

        },

        Output_GetBulbPowerStateInverted: function (ep) {
            var target_mac_or_name = ep[2];
            // Find the bulb by name
            var bulb = getBulbByName(target_mac_or_name);
            if (!bulb) {
                // Find the bulb by MAC address
                bulb = getBulbByMac(target_mac_or_name);
            }
            if (!bulb) {
                CFM.log("GetBulbPowerState: Bulb not found: " + target_mac_or_name);
                return;
            }

            // Request the actual state of the bulb, if we get a reply the real output will be triggered after the current stored state is returned below
            sendBroadcastPacket(encodePacket(0x14, null, bulb.target_mac));

            // Return the bulb power state as a zero or one.
            return bulb.power ? 0 : 1;

        },

        Output_GetBulbDimLevel: function (ep) {
            var target_mac_or_name = ep[2];
            // Find the bulb by name
            var bulb = getBulbByName(target_mac_or_name);
            if (!bulb) {
                // Find the bulb by MAC address
                bulb = getBulbByMac(target_mac_or_name);
            }
            if (!bulb) {
                CFM.log("GetBulbDimLevel: Bulb not found: " + target_mac_or_name);
                return;
            }

            // Request the actual state of the bulb, if we get a reply the real output will be triggered after the current stored state is returned below
            sendBroadcastPacket(encodePacket(0x65, null, bulb.target_mac));

            // Return the bulb power state as a zero or one.
            return bulb.bright;

        },

        Output_GetBulbDimPercent: function (ep) {
            var target_mac_or_name = ep[2];
            // Find the bulb by name
            var bulb = getBulbByName(target_mac_or_name);
            if (!bulb) {
                // Find the bulb by MAC address
                bulb = getBulbByMac(target_mac_or_name);
            }
            if (!bulb) {
                CFM.log("GetBulbDimPercent: Bulb not found: " + target_mac_or_name);
                return;
            }

            // Request the actual state of the bulb, if we get a reply the real output will be triggered after the current stored state is returned below
            sendBroadcastPacket(encodePacket(0x65, null, bulb.target_mac));

            // Return the bulb power state as a zero or one.
            return Math.floor((100/65535)*bulb.bright) + "%";
        }
	};
};