var com_commandfusion_sip = function(instanceID, instanceSettings, moduleDef) {


	/* ----------------------------------------------------
	 * MODULE PRIVATE MEMBERS
	 * ----------------------------------------------------
	 */

	// Get the name of the module from the instance settings.
	// Use this for your output logging to easily identify your module output logs when debugging.
	// By default debugging logs will be prefixed with the instance ID.
	var moduleName = instanceSettings.name,
		dialogFormat = "json",
		enteredNumber = "",
		makingCall = false,
		activeCalls = [],
		systemName = instanceID + ".SIP",
		/**
		 * Project data provided by Control Center on project launch, via function @Config_Startup
		 * @type {{token: string, serverURL: string, settingsSaveURL: string, logEventURL: string}}
		 */
		projectData = {
			token: "",
			serverURL: "http://cc.commandfusion.com",
			settingsSaveURL: "",
			logEventURL: "",
			zones: {}
		},
		/**
		 * Store any settings for the module here. This object will be converted to JSON string and saved to Control Center cloud
		 * when the function @Config_SaveSettings is called.
		 * @type {Object}
		 */
		settings = {
			devices: [], // Array of devices with their username and password mappings to allow multiple devices to run
						 // a single project, each with their own login info stored.
			addressbook: [] // Array of contacts, with keys 'name' and 'number'.
		};

	/* ----------------------------------------------------
	 * INTERNAL PRIVATE FUNCTIONS - These can only be called by the module code itself.
	 * ----------------------------------------------------
	 */

	function processChar(char) {
		switch (char) {
			case "dot":
				char = ".";
				break;
			case "asterisk":
			case "star":
				char = "*";
				break;
			case "plus":
				char = "+";
				break;
			case "hash":
				char = "#";
				break;
		}
		return char;
	}

	function onlyUnique(value, index, self) {
		return self.indexOf(value) === index;
	}

	function processIncomingData(fbName, data) {
		var msg;
		CFM.triggerOutput([instanceID, "msg"], data);
		try {
			data = JSON.parse(data);
		} catch (e) {
			msg = "Error processing incoming data: " + e.message;
			CF.log(msg);
			CFM.triggerOutput([instanceID, 'msg'], msg);
			return;
		}

		var dataCommand = data[0];
		data = data[1];

		switch (dataCommand) {
			case "STATUS":

				CFM.triggerOutput([instanceID, "call-status"], data.state);
				CFM.triggerOutput([instanceID, "active-number"], data.remote);

				// Manage active call list
				if (data.hasOwnProperty('id')) {
					// Remove any previous entry of the call ID
					activeCalls = activeCalls.filter(function(el) {
						return el.id != data.id;
					});
					// Add the new call state back to the array
					activeCalls.push(data);
				}

				if (data.state == "CLOSED") {
					CFM.triggerOutput([instanceID, "call-status-msg"], "Call Closed");
					CFM.triggerOutput([instanceID, "incoming"], "0");
					CFM.triggerOutput([instanceID, "outgoing"], "0");
					CFM.triggerOutput([instanceID, "in-call"], "0");
					CFM.triggerOutput([instanceID, "mute-state"], "0");
					CFM.triggerOutput([instanceID, "hold-state"], "0");
					hidePopup(instanceID + "_popup_incoming");
					hidePopup(instanceID + "_popup_incall_audio");
					hidePopup(instanceID + "_popup_incall_video");
					hidePopup(instanceID + "_popup_starting");
					CFM.triggerOutput([instanceID, "notification"], JSON.stringify({type: "none", hideNotifications: true}));
					if (makingCall) {
						showPopup(instanceID + "_popup_failed");
					}
					// Remove from active calls list
					activeCalls = activeCalls.filter(function(el) {
						return el.id != data.id;
					});

					makingCall = false;
				} else if (data.state == "ESTABLISHED") {
					CFM.triggerOutput([instanceID, "call-status-msg"], "Active Call");
					CFM.triggerOutput([instanceID, "mute-state"], data.mute ? "1" : "0");
					CFM.triggerOutput([instanceID, "hold-state"], data.hold ? "1" : "0");
					CFM.triggerOutput([instanceID, "incoming"], "0");
					CFM.triggerOutput([instanceID, "outgoing"], "0");
					CFM.triggerOutput([instanceID, "in-call"], "1");
					hidePopup(instanceID + "_popup_incoming");
					hidePopup(instanceID + "_popup_starting");
					hidePopup(instanceID + "_popup_failed");
					// TODO - check if call has video?
					showPopup(instanceID + "_popup_incall_video");
					showInCallNotification();
					makingCall = false;
				} else if (data.state == "RINGING") {
					CFM.triggerOutput([instanceID, "incoming"], "0");
					CFM.triggerOutput([instanceID, "in-call"], "0");
					CFM.triggerOutput([instanceID, "outgoing"], "1");
					CFM.triggerOutput([instanceID, "call-status-msg"], "Ringing...");
					hidePopup(instanceID + "_popup_incall_audio");
					hidePopup(instanceID + "_popup_incall_video");
					if (data.incoming) {
						hidePopup(instanceID + "_popup_starting");
						// TODO - check if call has video?
						showPopup(instanceID + "_popup_incoming");
					} else {
						hidePopup(instanceID + "_popup_incoming");
						showPopup(instanceID + "_popup_starting");
					}
				} else if (data.state == "UNKNOWN") {
					makingCall = false;
					CFM.triggerOutput([instanceID, "call-status-msg", "Unknown"]);
					hidePopup(instanceID + "_popup_incoming");
					hidePopup(instanceID + "_popup_incall_audio");
					hidePopup(instanceID + "_popup_incall_video");
					CFM.triggerOutput([instanceID, "notification"], JSON.stringify({type: "none", hideNotifications: true}));
				} else if (data.state == "TRYING") {
					CFM.triggerOutput([instanceID, "incoming"], "0");
					CFM.triggerOutput([instanceID, "in-call"], "0");
					CFM.triggerOutput([instanceID, "outgoing"], "1");
					CFM.triggerOutput([instanceID, "call-status-msg"], "Attempting Call...");
					hidePopup(instanceID + "_popup_incoming");
					hidePopup(instanceID + "_popup_incall_audio");
					hidePopup(instanceID + "_popup_incall_video");
					showPopup(instanceID + "_popup_starting");
				}
				break;
			case "INCOMING":
				// Now handled by the JS watch

				// // Manage active call list
				// data["state"] = dataCommand;
				// data["remote"] = data.caller || "";
				//
				// CF.log("DATA:");
				// CF.logObject(data);
				//
				// if (data.hasOwnProperty('id')) {
				// 	// Remove any previous entry of the call ID
				// 	activeCalls = activeCalls.filter(function(el) {
				// 		return el.id != data.id;
				// 	});
				// 	// Add the new call state back to the array
				// 	activeCalls.push(data);
				// }
				//
				// CF.log("ACTIVE CALLS:");
				// CF.logObject(activeCalls);
				//
				// CFM.triggerOutput([instanceID, "outgoing"], "0");
				// CFM.triggerOutput([instanceID, "incoming"], "1");
				// CFM.triggerOutput([instanceID, "active-number"], data.caller);
				// hidePopup(instanceID + "_popup_incall_audio");
				// hidePopup(instanceID + "_popup_incall_video");
				// hidePopup(instanceID + "_popup_starting");
				// showPopup(instanceID + "_popup_incoming");
				break;
			case "OUTGOING":
				data["state"] = dataCommand;
				// Manage active call list
				if (data.hasOwnProperty('id')) {
					// Remove any previous entry of the call ID
					activeCalls = activeCalls.filter(function(el) {
						return el.id != data.id;
					});
					// Add the new call state back to the array
					activeCalls.push(data);
				}
				CF.log("OUTGOING CALL TO: " + enteredNumber);

				CFM.triggerOutput([instanceID, "incoming"], "0");
				CFM.triggerOutput([instanceID, "outgoing"], "1");
				CFM.triggerOutput([instanceID, "dial-number"], enteredNumber);
				CFM.triggerOutput([instanceID, "active-number"], enteredNumber);
				hidePopup(instanceID + "_popup_incoming");
				hidePopup(instanceID + "_popup_incall_audio");
				hidePopup(instanceID + "_popup_incall_video");
				showPopup(instanceID + "_popup_starting");
				break;
			case "REGISTER":
				if (data.hasOwnProperty('status')) {
					if (data.status == "401") {
						CFM.triggerOutput([instanceID, "registered"], "0");
					} else if (data.status == "403") {
						CFM.triggerOutput([instanceID, "registered"], "0");
					} else if (data.status == "200") {
						CFM.triggerOutput([instanceID, "registered"], "1");
					}
				}
				break;
			case "ERROR":
				if (data.indexOf("CALL failed") !== -1) {
					// CALL FAILED error
					hidePopup(instanceID + "_popup_incoming");
					hidePopup(instanceID + "_popup_incall_audio");
					hidePopup(instanceID + "_popup_incall_video");
					hidePopup(instanceID + "_popup_starting");
					showPopup(instanceID + "_popup_failed");
				}
				break;
		}
	}

	function send (obj) {
		CFM.log("SENDING SIP COMMAND: " + JSON.stringify(obj));
		CF.send(systemName, JSON.stringify(obj));
	}

	function showPopup(popupTag) {
		CF.setProperties({join: popupTag, opacity: 1.0}, 0.0, 0.2);
	}

	function hidePopup(popupTag) {
		CF.setProperties({join: popupTag, opacity: 0.0}, 0.0, 0.2);
	}

	function sendToCallID(command, callID) {
		if (callID && callID != "0") {
			send([command, callID]);
		} else {
			if (activeCalls.length) {
				send([command, { id: activeCalls[activeCalls.length - 1].id }]);
			} else {
				send([command, {}]);
			}
		}
	}

	function getCall(id) {
		if (!id) return false;

		return activeCalls.find(function(call) {
			return call.id == id;
		});
	}

	function showInCallNotification() {
		CF.logObject(activeCalls);
		var lastCall = activeCalls[activeCalls.length - 1];
		CFM.triggerOutput([instanceID, "notification"], JSON.stringify({
			type: "info", title: "In Call", msg: lastCall.remote ? lastCall.remote : "Unknown call",
			primary: {
				event: {
					"action-id": "module-input",
					"module": instanceID,
					"input": "resume"

				},
				text: "RETURN TO CALL",
				type: "default"
			},
			secondary: {
				event: {
					"action-id": "module-input",
					"module": instanceID,
					"input": "hangup"

				},
				text: "HANGUP",
				type: "danger"
			},
			close: false
		}));
	}

	function uniqueID() {
		function chr4() {
			return Math.random().toString(16).slice(-4);
		}
		return chr4() + chr4() +
			'-' + chr4() +
			'-' + chr4() +
			'-' + chr4() +
			'-' + chr4() + chr4() + chr4();
	}

	function generateListItem(contact) {
		return {
			s1: contact.name,
			s2: contact.number,
			s3: contact.id,
			d1: {
				tokens: contact
			},
			d2: {
				tokens: contact
			},
			d3: {
				tokens: contact
			}
		};
	}

	function RefreshAddressBook(showID) {
		var listData = [], listTag = "l" + instanceID + "_addressbook", showIndex;
		CF.listRemove(listTag);
		settings.addressbook.forEach(function(contact) {
			listData.push(generateListItem(contact));
			if (showID && contact.id == showID) {
				showIndex = listData.length;
			}
		});

		if (listData.length) {
			CF.listAdd(listTag, listData);
		}

		// Ensure a specific contact ID is visible if requested
		if (showIndex) {
			CF.listScroll(listTag, showIndex, CF.VisiblePosition, true);
		}
	}
	
	function getAddressBookContact(id) {
		if (!id) return false;

		return settings.addressbook.find(function(contact) {
			return contact.id == id;
		});
	}

	return {
		/**
		 * Observe any parameter change at runtime.
		 * This is NOT called on startup. See the @onStartup function for how to process the module parameters on startup.
		 * @param parameter {String} - The name of the parameter that has changed.
		 * @param value {String} - The new value of the parameter.
		 */
		onParameterChange: function(parameter, oldvalue, value) {
			CFM.log(moduleName + " onParameterChange: " + parameter + " = " + value + (" (previous value: " + oldvalue + ")"));
			// Then do something with the parameter data, like store it for later in a private variable defined above
			// Or store it in the settings object so that it will persist across project updates.
			switch (parameter) {
				case "user":
					CF.setSystemProperties(systemName, {user: value});
					// Assign the username to this device in the module settings
					var assigned = false;
					if (settings.devices) {
						settings.devices.forEach(function(device) {
							if (device.uuid && device.uuid == CF.device.uuid) {
								device.user = value;
								assigned = true;
							}
						});
					}
					if (!assigned) {
						settings.devices.push({uuid: CF.device.uuid, user: value});
					}
					this.Config_SaveSettings();
					break;
				case "password":
					CF.setSystemProperties(systemName, {password: value});
					// Assign the password to this device in the module settings
					var assigned = false;
					settings.devices.forEach(function(device) {
						if (device.uuid && device.uuid == CF.device.uuid) {
							device.password = value;
							assigned = true;
						}
					});
					if (!assigned) {
						settings.devices.push({uuid: CF.device.uuid, password: value});
					}
					this.Config_SaveSettings();
					break;
				case "identity":
					CF.setSystemProperties(systemName, {identity: value});
					break;
				case "dtmf-mode":
					CF.setSystemProperties(systemName, {DTMFMode: parseInt(value, 10)});
					break;
				case "transport-type":
					CF.setSystemProperties(systemName, {transport: value});
					break;
				case "rtp-encryption":
					CF.setSystemProperties(systemName, {RTPEncryption: parseInt(value, 10)});
					break;
				//case "dialog-format":
				//	CF.setSystemProperties(systemName, {dialogFormat: value});
				//	dialogFormat = value;
				//	break;
			}
		},

		/**
		 * This function is called as soon as the app is launched and the JS framework is ready.
		 * Any setup code is called from here.
		 */
		onStartup: function() {
			CFM.log(moduleName + " onStartup");

			// Setup watch for processing incoming data from SIP
			CF.watch(CF.FeedbackMatchedEvent, systemName, "SIP_FB", processIncomingData);

			// Force update from all module parameters on startup
			var props = { enabled: true };
			for (var key in CFM.parameters) {
				var value = CFM.parameters[key];
				if (value == "") continue; // Skip empty params (use defaults in SIP engine)
				switch (key) {
					case "user":
						props['user'] = value;
						break;
					case "password":
						props['password'] = value;
						break;
					case "identity":
						props['identity'] = value;
						break;
					case "dtmf-mode":
						props['DTMFMode'] = parseInt(value, 10);
						break;
					case "transport-type":
						props['transport'] = value;
						break;
					case "rtp-encryption":
						props['RTPEncryption'] = parseInt(value, 10);
						break;
					//case "dialog-format":
					//	props['dialogFormat'] = value;
					//	dialogFormat = value;
					//	break;
				}
			}

			// Override any username and password config with data stored in settings for this specific device:
			settings.devices.forEach(function(device) {
				if (device.uuid && device.uuid == CF.device.uuid) {
					props['user'] = device.user || props['user'];
					props['password'] = device.password || props['password'];

					// Update the config layout to show the auth details used from settings
					CF.setJoins([{join: instanceID + "_input_user", value: props['user']}, {join: instanceID + "_input_password", value: props['password']}]);
				}
			});

			CF.setSystemProperties(systemName, props);

			// Listen for SIP events
			CF.watch(CF.SIPCallEvent, CF.SIPCallRingingIncoming, function(event, system, callID, oppositeNumber) {
				CF.log("SIPCallEvent event="+event+" system="+system+", callID="+callID+", number="+oppositeNumber);

				if (system != systemName) {
					CF.log("SIP event not for this module!");
					return;
				}

				// Manage active call list
				if (callID) {
					// Remove any previous entry of the call ID
					activeCalls = activeCalls.filter(function(el) {
						return el.id != callID;
					});
					// Add the new call state back to the array
					activeCalls.push({id: callID, caller: oppositeNumber, remote: oppositeNumber, state: "INCOMING"});
				}

				CFM.triggerOutput([instanceID, "outgoing"], "0");
				CFM.triggerOutput([instanceID, "incoming"], "1");
				CFM.triggerOutput([instanceID, "active-number"], oppositeNumber);
				hidePopup(instanceID + "_popup_incall_audio");
				hidePopup(instanceID + "_popup_incall_video");
				hidePopup(instanceID + "_popup_starting");
				showPopup(instanceID + "_popup_incoming");
			});

			RefreshAddressBook();
		},

		/* ----------------------------------------------------
		 * MODULE INPUTS (SETTERS) - These are called via module routes
		 * Function names should match the input 'setter' params in the manifest.json
		 * ----------------------------------------------------
		 */

		Input_AppendNumber: function(ep, char) {
			char = processChar(char);
			enteredNumber = "" + enteredNumber + char;
			CFM.triggerOutput([instanceID, 'dial-number'], enteredNumber);
		},

		Input_DeleteNumber: function(ep) {
			enteredNumber = enteredNumber.slice(0, -1);
			CFM.triggerOutput([instanceID, 'dial-number'], enteredNumber);
		},

		Input_ClearNumber: function(ep) {
			enteredNumber = "";
			CFM.triggerOutput([instanceID, 'dial-number'], "");
		},

		Input_DTMF: function(ep, char) {
			char = processChar(char);
			enteredNumber += char;
			CFM.triggerOutput([instanceID, 'dial-number'], enteredNumber);
			if ((ep.length < 3 || !ep[2])) {
				// Default to only active call, or 0 if no active calls
				ep[2] = activeCalls.length ? activeCalls[activeCalls.length - 1].id: 0;
			}
			if (ep[2] && ep[2] != "0") {
				send(["SEND-DTMF", { id: ep[2], tones: char }]);
			} else {
				send(["SEND-DTMF", {tones: char}]);
			}
		},

		Input_Call: function(ep, number) {
			enteredNumber = number || enteredNumber;
			makingCall = true;
			if (ep.length < 3 || !ep[2]) {
				ep[2] = "0"; // Default to audio only call
			}
			if (ep[2] != "0") {
				var tag = instanceID + "_SIPVideo"; // Default tag name for SIP Video object if no tag is given, but video call is being attempted
				if (ep.length > 3) {
					tag = ep[3]; // Using a custom SIP Video object tag
				}
				if (dialogFormat == "json") {
					send(["CALL", {remote: enteredNumber, avObject: tag}]); // Video call
				} else {
					CF.send(systemName, "CALL " + enteredNumber + " " + tag);
				}
			} else {
				if (dialogFormat == "json") {
					send(["CALL", { remote: enteredNumber }]); // Audio Only call
				} else {
					CF.send(systemName, "CALL " + enteredNumber + " SIP-AUDIO-ONLY");
				}
			}
		},

		Input_Decline: function(ep, callID) {
			if (callID) {
				send(["DECLINE", callID]);
			} else {
				if (activeCalls.length) {
					send(["DECLINE", { id: activeCalls[activeCalls.length - 1].id }]);
				} else {
					send(["DECLINE", {}]);
				}
			}
			hidePopup(instanceID + "_popup_incoming");
		},

		Input_Answer: function(ep, callID) {
			if (ep.length < 3 || !ep[2]) {
				ep[2] = "0"; // Default to audio only call
			}
			var tag = "";
			if (ep[2] != "0" && ep.length > 3) {
				tag = ep[3]; // Using a custom SIP Video object tag
			} else {
				tag = (ep[2] == "0" ? "SIP-AUDIO-ONLY" : instanceID + "_SIPVideo"); // Default tag name for SIP Video object if no tag is given, but video call is being attempted
			}
			if (dialogFormat == "json") {
				if (callID) {
					send(["ACCEPT", { id: callID, avObject: tag }]);
				} else {
					if (activeCalls.length) {
						send(["ACCEPT", { id: activeCalls[activeCalls.length - 1].id, avObject: tag }]);
					} else {
						send(["ACCEPT", { avObject: tag }]);
					}
				}
			} else {
				if (callID) {
					CF.send(systemName, "ACCEPT " + callID + " " + tag);
				} else {
					if (activeCalls.length) {
						CF.send(systemName, "ACCEPT " + activeCalls[activeCalls.length - 1].id + " " + tag);
					} else {
						CF.send(systemName, "ACCEPT " + tag);
					}
				}
			}

			//hidePopup(instanceID + "_popup_incoming");
			// Show correct incall popup depending on audio or video call
			//showPopup(instanceID + "_popup_incall_" + (tag == "SIP-AUDIO-ONLY" ? "audio" : "video"));
		},

		Input_Hangup: function(ep, callID) {
			if (callID) {
				if (dialogFormat == "json") {
					send(["HANGUP", { id: callID }]);
				} else {
					CF.send(systemName, "HANGUP " + callID);
				}
			} else {
				if (activeCalls.length) {
					if (dialogFormat == "json") {
						send(["HANGUP", { id: activeCalls[activeCalls.length - 1].id }]);
					} else {
						CF.send(systemName, "HANGUP " + activeCalls[activeCalls.length - 1].id);
					}
				} else {
					if (dialogFormat == "json") {
						send(["HANGUP", {}]);
					} else {
						CF.send(systemName, "HANGUP");
					}
				}
			}

			hidePopup(instanceID + "_popup_incoming");
			hidePopup(instanceID + "_popup_starting");
			hidePopup(instanceID + "_popup_incall_video");
			hidePopup(instanceID + "_popup_incall_audio");
		},

		Input_Mute: function(ep, callID) {
			sendToCallID("MUTE", callID);
		},

		Input_MuteToggle: function(ep, callID) {
			var theCall = getCall(callID);
			if (!theCall) {
				if (activeCalls.length) {
					theCall = activeCalls[activeCalls.length - 1];
				} else {
					CF.log("Mute toggle failed. No active calls could be found.");
					return;
				}
			}
			sendToCallID(theCall.hold ? "UNMUTE" : "MUTE", theCall.id);
		},

		Input_Unmute: function(ep, callID) {
			sendToCallID("UNMUTE", callID);
		},

		Input_Hold: function(ep, callID) {
			sendToCallID("HOLD", callID);
		},

		Input_HoldToggle: function(ep, callID) {
			var theCall = getCall(callID);
			if (!theCall) {
				if (activeCalls.length) {
					theCall = activeCalls[activeCalls.length - 1];
				} else {
					CF.log("Hold toggle failed. No active calls could be found.");
					return;
				}
			}
			sendToCallID(theCall.hold ? "UNHOLD" : "HOLD", theCall.id);
		},

		Input_Unhold: function(ep, callID) {
			sendToCallID("UNHOLD", callID);
		},

		Input_ResumeUI: function(ep) {
			if (activeCalls && activeCalls.length) {
				// Make sure the in-call notification is shown again because it will automatically hide if a
				// notification button was pressed to call this event
				showInCallNotification();
				// TODO - use call video state to determine which popup to show
				// Default to show video one for now
				showPopup(instanceID + "_popup_incall_video");
				//showPopup(instanceID + "_popup_incall_audio");
			}
		},
		Input_AddressBookAdd: function(ep, number) {
			// Validate entry
			if (number == "undefined" || number == "") {
				CFM.log("Could not add address book entry: Missing contact number.");
				return;
			}
			if (ep.length < 3 || ep[2] == "") {
				CFM.log("Could not add address book entry: Missing contact name.");
				return;
			}
			
			var id = uniqueID();
			settings.addressbook.push({id: id, name: ep[2], number: number});

			RefreshAddressBook(id);

			this.Config_SaveSettings();
		},

		Input_AddressBookEdit: function(ep, number) {
			// Validate entry
			if (number == "undefined" || number == "") {
				CFM.log("Could not edit address book entry: Missing contact number.");
				return;
			}
			if (ep.length < 4 || ep[2] == "") {
				CFM.log("Could not add address book entry: Missing contact ID.");
				return;
			}
			if (ep.length < 4 || ep[3] == "") {
				CFM.log("Could not add address book entry: Missing contact name.");
				return;
			}
			var id = ep[2];
			var contact = getAddressBookContact(id);
			if (!contact){
				CFM.log("Could not edit address book entry: Contact ID not found - " + id);
				return;
			}

			var listTag = "l" + instanceID + "_addressbook";

			var that = this;

			// Update the item within the list without doing a full list refresh
			// Find the index of the item in the list by its ID (which is stored in a list items' button tokens)
			CF.listContents(listTag, 0, 0, function(list) {
				var updateIndex = list.findIndex(function(contact) {
					// Check the list item has a digital join 1, and tokens within that join and the ID token matches
					return !!contact.d1 && !!contact.d1.tokens && contact.d1.tokens["id"] == id;
				});

				if (updateIndex !== -1) {
					var contact = {
						id: id,
						name: ep[3],
						number: number
					};

					// Update the item in settings
					settings.addressbook[updateIndex] = contact;

					// Update the item in the list
					var listItem = generateListItem(contact);
					listItem["index"] = updateIndex;
					CF.listUpdate(listTag, [listItem]);

					that.Config_SaveSettings();
				}
			});
		},

		Input_AddressBookDelete: function(ep, id) {
			// Validate entry
			var contact = getAddressBookContact(id);
			if (!contact){
				CFM.log("Could not delete address book entry: Contact ID not found - " + id);
				return;
			}

			var listTag = "l" + instanceID + "_addressbook";

			var that = this;

			// Remove the item from the list without doing a full list refresh
			// Find the index of the item in the list by its ID (which is stored in a list items' button tokens)
			CF.listContents(listTag, 0, 0, function(list) {
				var removeIndex = list.findIndex(function(contact) {
					// Check the list item has a digital join 1, and tokens within that join and the ID token matches
					return !!contact.d1 && !!contact.d1.tokens && contact.d1.tokens["id"] == id;
				});

				if (removeIndex !== -1) {
					// Delete the item from settings
					settings.addressbook.splice(removeIndex, 1);
					// Remove the item from the list
					CF.listRemove(listTag, removeIndex, 1);

					that.Config_SaveSettings();
				}
			});
		},

		/**
		 * Save Settings to the Control Center cloud. The settings can then be persisted across project updates.
		 * @param ep {Array} - Route endpoint data consisting of the module instance ID and the module input ID 'config-save-settings'.
		 * @param settingsSaveURL {String} - The URL to push the saved settings to
		 */
		Config_SaveSettings: function(ep, settingsSaveURL) {
			projectData.settingsSaveURL = settingsSaveURL || projectData.settingsSaveURL;
			// Convert the module settings to JSON string
			var settingsString = JSON.stringify(settings);

			// Save the settings into the local global token first
			CF.getJoin(CF.GlobalTokensJoin, function(j,v,tokens) {
				if (typeof tokens["module_settings"] !== "undefined") {
					// Read the current settings
					var settingsData = JSON.parse(tokens["module_settings"]);
					// Update just the settings for the current module
					settingsData[instanceID] = JSON.parse(settingsString);
					// Save the settings back to the global token
					CF.setToken(CF.GlobalTokensJoin, "module_settings", JSON.stringify(settingsData));
				}
			});

			// Save settings back to the cloud
			if (!projectData.settingsSaveURL) {
				CF.log(moduleName + " SAVE SETTINGS: settings URL was not defined. Settings could not be saved.");
				return;
			}
			CF.request(projectData.settingsSaveURL,
				"POST",
				null,
				{'settings': settingsString, 'module_instance': instanceID},
				function(status, headers, body) {
				}
			);
		},

		/**
		 * Load Settings as defined from the Control Center cloud at the time the project archive was built. This is called after startup.
		 * @param ep {Array} - Route endpoint data consisting of the module instance ID and the module input ID 'config-load-settings'.
		 * @param settingsData {String} - JSON formatted String of settings for the module to process.
		 */
		Config_LoadSettings: function(ep, settingsData) {
			// Convert the settings data to an object
			try {
				settings = JSON.parse(settingsData);

				RefreshAddressBook();

				// Reauth based on stored settings
				var props = {};

				// Override any username and password config with data stored in settings for this specific device:
				if (settings.devices) {
					settings.devices.forEach(function(device) {
						if (device.uuid && device.uuid == CF.device.uuid) {
							props['user'] = device.user;
							props['password'] = device.password;

							// Update the config layout to show the auth details used from settings
							CF.setJoins([{join: instanceID + "_input_user", value: device.user}, {join: instanceID + "_input_password", value: device.password}]);
						}
					});
				}

				CF.setSystemProperties(systemName, props);
			} catch (e) {
				CF.log(moduleName + " ERROR LOADING SETTINGS: " + e.stack);
			}
		},

		/**
		 * When project is launched, this function is called with the project data.
		 * @param ep {Array} - Route Endpoint data that was used to call this input.
		 * @param projectInfo {String} - JSON formatted String of project settings such as URLs and project token.
		 */
		Config_Startup: function(ep, projectInfo) {
			try {
				projectData = JSON.parse(projectInfo);
			} catch (e) {
				CF.log(moduleName + " ERROR PROCESSING PROJECT INFO: " + e.stack);
			}
		}
	};
};
